"""
Test Python 3.8-specific features.

For example, one such feature is walrus operator used in named expressions.
We have to exclude these tests running on prior versions of Python since the syntax would be considered
invalid.
"""

import sys

if sys.version_info < (3, 8):

    def load_tests(loader, suite, pattern):  # pylint: disable=unused-argument
        """Ignore all the tests for lower Python versions."""
        return suite



"""
Test Python 3.-specific features.

For example, one such feature is decorator ``dataclasses.dataclass``.
"""

import sys

if sys.version_info < (3, 7):

    def load_tests(loader, suite, pattern):  # pylint: disable=unused-argument
        """Ignore all the tests for lower Python versions."""
        return suite



# The module ``unittest`` supports async only from 3.8 on.
# That is why we had to move this test to 3.8 specific tests.

# pylint: disable=missing-docstring, invalid-name, no-member, unnecessary-lambda
import unittest
from typing import Optional, List

import icontract


class TestSyncFunctionAsyncConditionFail(unittest.IsolatedAsyncioTestCase):
    def test_precondition(self) -> None:
        async def x_greater_zero(x: int) -> bool:
            return x > 0

        @icontract.require(x_greater_zero)
        def some_func(x: int) -> int:
            return x * 10

        value_error = None  # type: Optional[ValueError]
        try:
            _ = some_func(100)
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertRegex(
            str(value_error),
            r'^Unexpected coroutine \(async\) condition <.*> for a sync function <.*\.some_func at .*>.')

    def test_postcondition(self) -> None:
        async def result_greater_zero(result: int) -> bool:
            return result > 0

        @icontract.ensure(result_greater_zero)
        def some_func() -> int:
            return 100

        value_error = None  # type: Optional[ValueError]
        try:
            _ = some_func()
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertRegex(
            str(value_error),
            r'^Unexpected coroutine \(async\) condition <.*> for a sync function <.*\.some_func at .*>.')

    def test_snapshot(self) -> None:
        async def capture_len_lst(lst: List[int]) -> int:
            return len(lst)

        @icontract.snapshot(capture_len_lst, name="len_lst")
        @icontract.ensure(lambda OLD, lst: OLD.len_lst + 1 == len(lst))
        def some_func(lst: List[int]) -> None:
            lst.append(1984)

        value_error = None  # type: Optional[ValueError]
        try:
            some_func([1])
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertRegex(
            str(value_error), r'^Unexpected coroutine \(async\) snapshot capture <function .*\.capture_len_lst at .*> '
            r'for a sync function <function .*\.some_func at .*>\.')


class TestSyncFunctionConditionCoroutineFail(unittest.IsolatedAsyncioTestCase):
    def test_precondition(self) -> None:
        async def x_greater_zero(x: int) -> bool:
            return x > 0

        @icontract.require(lambda x: x_greater_zero(x))
        def some_func(x: int) -> int:
            return x * 10

        value_error = None  # type: Optional[ValueError]
        try:
            _ = some_func(100)
        except ValueError as err:
            value_error = err

        assert value_error is not None

        self.assertRegex(
            str(value_error),
            r"^Unexpected coroutine resulting from the condition <function .*> for a sync function <function .*>\.$")

    def test_postcondition(self) -> None:
        async def result_greater_zero(result: int) -> bool:
            return result > 0

        @icontract.ensure(lambda result: result_greater_zero(result))
        def some_func() -> int:
            return 100

        value_error = None  # type: Optional[ValueError]
        try:
            _ = some_func()
        except ValueError as err:
            value_error = err

        assert value_error is not None

        self.assertRegex(
            str(value_error),
            r"^Unexpected coroutine resulting from the condition <function .*> for a sync function <function .*>\.$")

    def test_snapshot(self) -> None:
        async def capture_len_lst(lst: List[int]) -> int:
            return len(lst)

        @icontract.snapshot(lambda lst: capture_len_lst(lst), name="len_lst")
        @icontract.ensure(lambda OLD, lst: OLD.len_lst + 1 == len(lst))
        def some_func(lst: List[int]) -> None:
            lst.append(1984)

        value_error = None  # type: Optional[ValueError]
        try:
            some_func([1])
        except ValueError as err:
            value_error = err

        assert value_error is not None
        self.assertRegex(
            str(value_error), r'^Unexpected coroutine resulting '
            r'from the snapshot capture <function .*> of a sync function <function .*>.$')


class TestAsyncInvariantsFail(unittest.IsolatedAsyncioTestCase):
    def test_that_async_invariants_reported(self) -> None:
        async def some_async_invariant(self: 'A') -> bool:
            return self.x > 0

        value_error = None  # type: Optional[ValueError]
        try:
            # pylint: disable=unused-variable
            @icontract.invariant(some_async_invariant)
            class A:
                def __init__(self) -> None:
                    self.x = 100
        except ValueError as error:
            value_error = error

        assert value_error is not None

        self.assertEqual(
            "Async conditions are not possible in invariants as sync methods such as __init__ have to be wrapped.",
            str(value_error))


if __name__ == '__main__':
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=unused-argument
# pylint: disable=no-member
import textwrap
import unittest
from typing import NamedTuple, Optional  # pylint: disable=unused-import

import icontract
import tests.error


class TestOK(unittest.TestCase):
    def test_on_named_tuple(self) -> None:
        # This test is related to the issue #171.
        #
        # The test could not be executed under Python 3.6 as the ``inspect`` module
        # could not figure out the type of getters.
        @icontract.invariant(lambda self: self.first > 0)
        class RightHalfPlanePoint(NamedTuple):
            first: int
            second: int

        _ = RightHalfPlanePoint(1, 0)

        self.assertEqual('Create new instance of RightHalfPlanePoint(first, second)',
                         RightHalfPlanePoint.__new__.__doc__)


class TestViolation(unittest.TestCase):
    def test_on_named_tuple(self) -> None:
        # This test is related to the issue #171.
        #
        # The test could not be executed under Python 3.6 as the ``inspect`` module
        # could not figure out the type of getters.
        @icontract.invariant(lambda self: self.second > 0)
        @icontract.invariant(lambda self: self.first > 0)
        class RightHalfPlanePoint(NamedTuple):
            first: int
            second: int

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = RightHalfPlanePoint(1, -1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual(
            textwrap.dedent('''\
                self.second > 0:
                self was RightHalfPlanePoint(first=1, second=-1)
                self.second was -1'''), tests.error.wo_mandatory_location(str(violation_error)))


if __name__ == '__main__':
    unittest.main()



"""
Test Python 3.6-specific features.

For example, one such feature is literal string interpolation.
"""

import sys

if sys.version_info < (3, 6):
    def load_tests(loader, suite, pattern):  # pylint: disable=unused-argument
        """Ignore all the tests for lower Python versions."""
        return suite



# pylint: disable=missing-docstring
# pylint: disable=no-self-use
# pylint: disable=unnecessary-lambda
import unittest
from typing import List

import icontract


class TestPrecondition(unittest.IsolatedAsyncioTestCase):
    async def test_ok(self) -> None:
        order = []  # type: List[str]

        @icontract.require(lambda: another_func())
        @icontract.require(lambda: yet_another_func())
        async def some_func() -> bool:
            order.append(some_func.__name__)
            return True

        @icontract.require(lambda: some_func())
        @icontract.require(lambda: yet_yet_another_func())
        async def another_func() -> bool:
            order.append(another_func.__name__)
            return True

        async def yet_another_func() -> bool:
            order.append(yet_another_func.__name__)
            return True

        async def yet_yet_another_func() -> bool:
            order.append(yet_yet_another_func.__name__)
            return True

        await some_func()

        self.assertListEqual(['yet_another_func', 'yet_yet_another_func', 'some_func', 'another_func', 'some_func'],
                             order)

    async def test_recover_after_exception(self) -> None:
        order = []  # type: List[str]
        some_func_should_raise = True

        class CustomError(Exception):
            pass

        @icontract.require(lambda: another_func())  # pylint: disable=unnecessary-lambda
        @icontract.require(lambda: yet_another_func())  # pylint: disable=unnecessary-lambda
        async def some_func() -> bool:
            order.append(some_func.__name__)
            if some_func_should_raise:
                raise CustomError('some_func_should_raise')
            return True

        @icontract.require(lambda: some_func())  # pylint: disable=unnecessary-lambda
        @icontract.require(lambda: yet_yet_another_func())  # pylint: disable=unnecessary-lambda
        async def another_func() -> bool:
            order.append(another_func.__name__)
            return True

        async def yet_another_func() -> bool:
            order.append(yet_another_func.__name__)
            return True

        async def yet_yet_another_func() -> bool:
            order.append(yet_yet_another_func.__name__)
            return True

        try:
            await some_func()
        except CustomError:
            pass

        self.assertListEqual(['yet_another_func', 'yet_yet_another_func', 'some_func'], order)

        # Reset for the next experiment
        order = []
        some_func_should_raise = False

        await some_func()

        self.assertListEqual(['yet_another_func', 'yet_yet_another_func', 'some_func', 'another_func', 'some_func'],
                             order)


class TestPostcondition(unittest.IsolatedAsyncioTestCase):
    async def test_ok(self) -> None:
        order = []  # type: List[str]
        another_func_should_raise = True

        class CustomError(Exception):
            pass

        @icontract.ensure(lambda: another_func())  # pylint: disable=unnecessary-lambda
        @icontract.ensure(lambda: yet_another_func())  # pylint: disable=unnecessary-lambda
        async def some_func() -> bool:
            order.append(some_func.__name__)
            return True

        @icontract.ensure(lambda: some_func())  # pylint: disable=unnecessary-lambda
        @icontract.ensure(lambda: yet_yet_another_func())  # pylint: disable=unnecessary-lambda
        async def another_func() -> bool:
            order.append(another_func.__name__)
            if another_func_should_raise:
                raise CustomError('some_func_should_raise')

            return True

        async def yet_another_func() -> bool:
            order.append(yet_another_func.__name__)
            return True

        async def yet_yet_another_func() -> bool:
            order.append(yet_yet_another_func.__name__)
            return True

        try:
            await some_func()
        except CustomError:
            pass

        self.assertListEqual(['some_func', 'yet_another_func', 'another_func'], order)

        # Reset for the next experiments
        order = []
        another_func_should_raise = False

        await some_func()

        self.assertListEqual(['some_func', 'yet_another_func', 'another_func', 'yet_yet_another_func', 'some_func'],
                             order)

    async def test_recover_after_exception(self) -> None:
        order = []  # type: List[str]

        @icontract.ensure(lambda: another_func())  # pylint: disable=unnecessary-lambda
        @icontract.ensure(lambda: yet_another_func())  # pylint: disable=unnecessary-lambda
        async def some_func() -> bool:
            order.append(some_func.__name__)
            return True

        @icontract.ensure(lambda: some_func())  # pylint: disable=unnecessary-lambda
        @icontract.ensure(lambda: yet_yet_another_func())  # pylint: disable=unnecessary-lambda
        async def another_func() -> bool:
            order.append(another_func.__name__)
            return True

        async def yet_another_func() -> bool:
            order.append(yet_another_func.__name__)
            return True

        async def yet_yet_another_func() -> bool:
            order.append(yet_yet_another_func.__name__)
            return True

        await some_func()

        self.assertListEqual(['some_func', 'yet_another_func', 'another_func', 'yet_yet_another_func', 'some_func'],
                             order)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=missing-docstring,invalid-name,too-many-public-methods,no-self-use
# pylint: disable=unused-argument

import textwrap
import unittest
import math
from typing import Optional  # pylint: disable=unused-import

import icontract._represent
import tests.error
import tests.mock


class TestLiteralStringInterpolation(unittest.TestCase):
    def test_plain_string(self) -> None:
        @icontract.require(lambda x: f"something" == '')
        def func(x: float) -> float:
            return x

        violation_err = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=0)
        except icontract.ViolationError as err:
            violation_err = err

        self.assertIsNotNone(violation_err)
        self.assertEqual(
            'f"something" == \'\': f"something" was \'something\'',
            tests.error.wo_mandatory_location(str(violation_err)))

    def test_simple_interpolation(self) -> None:
        @icontract.require(lambda x: f"{x}" == '')
        def func(x: float) -> float:
            return x

        violation_err = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=0)
        except icontract.ViolationError as err:
            violation_err = err

        self.assertIsNotNone(violation_err)
        self.assertEqual(
            'f"{x}" == \'\': f"{x}" was \'0\'',
            tests.error.wo_mandatory_location(str(violation_err)))

    def test_string_formatting(self) -> None:
        @icontract.require(lambda x: f"{x!s}" == '')
        def func(x: float) -> float:
            return x

        violation_err = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1.984)
        except icontract.ViolationError as err:
            violation_err = err

        self.assertIsNotNone(violation_err)
        self.assertEqual(
            'f"{x!s}" == \'\': f"{x!s}" was \'1.984\'',
            tests.error.wo_mandatory_location(str(violation_err)))

    def test_repr_formatting(self) -> None:
        @icontract.require(lambda x: f"{x!r}" == '')
        def func(x: float) -> float:
            return x

        violation_err = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1.984)
        except icontract.ViolationError as err:
            violation_err = err

        self.assertIsNotNone(violation_err)
        self.assertEqual(
            'f"{x!r}" == \'\': f"{x!r}" was \'1.984\'',
            tests.error.wo_mandatory_location(str(violation_err)))

    def test_ascii_formatting(self) -> None:
        @icontract.require(lambda x: f"{x!a}" == '')
        def func(x: float) -> float:
            return x

        violation_err = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1.984)
        except icontract.ViolationError as err:
            violation_err = err

        self.assertIsNotNone(violation_err)
        self.assertEqual(
            'f"{x!a}" == \'\': f"{x!a}" was \'1.984\'',
            tests.error.wo_mandatory_location(str(violation_err)))

    def test_format_spec(self) -> None:
        @icontract.require(lambda x: f"{x:.3}" == '')
        def func(x: float) -> float:
            return x

        violation_err = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1.984)
        except icontract.ViolationError as err:
            violation_err = err

        self.assertIsNotNone(violation_err)
        self.assertEqual(
            'f"{x:.3}" == \'\': f"{x:.3}" was \'1.98\'',
            tests.error.wo_mandatory_location(str(violation_err)))

    def test_conversion_and_format_spec(self) -> None:
        @icontract.require(lambda x: f"{x!r:.3}" == '')
        def func(x: float) -> float:
            return x

        violation_err = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=1.984)
        except icontract.ViolationError as err:
            violation_err = err

        self.assertIsNotNone(violation_err)
        self.assertEqual(
            'f"{x!r:.3}" == \'\': f"{x!r:.3}" was \'1.9\'',
            tests.error.wo_mandatory_location(str(violation_err)))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# pylint: disable=missing-docstring,invalid-name,too-many-public-methods,no-self-use
# pylint: disable=unused-argument

import textwrap
import unittest
from typing import Optional  # pylint: disable=unused-import

import icontract._represent
import tests.error
import tests.mock


class TestReprValues(unittest.TestCase):
    def test_named_expression(self) -> None:
        @icontract.require(lambda x: (t := x + 1) and t > 1)  # pylint: disable=undefined-variable
        def func(x: int) -> int:
            return x

        violation_err = None  # type: Optional[icontract.ViolationError]
        try:
            func(x=0)
        except icontract.ViolationError as err:
            violation_err = err

        self.assertIsNotNone(violation_err)
        self.assertEqual(
            textwrap.dedent('''\
                (t := x + 1) and t > 1:
                t was 1
                x was 0'''), tests.error.wo_mandatory_location(str(violation_err)))


if __name__ == '__main__':
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=no-member

import unittest
from typing import Optional

import icontract

import tests.error


class TestAsyncMethod(unittest.IsolatedAsyncioTestCase):
    async def test_ok(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class A:
            def __init__(self) -> None:
                self.x = 100

            async def some_func(self) -> int:
                self.x = 200
                return self.x

            def another_func(self) -> int:
                self.x = 300
                return self.x

        a = A()
        result = await a.some_func()
        self.assertEqual(200, result)

        result = a.another_func()
        self.assertEqual(300, result)

    async def test_fail(self) -> None:
        @icontract.invariant(lambda self: self.x > 0)
        class A:
            def __init__(self) -> None:
                self.x = 100

            async def some_func(self) -> None:
                self.x = -1

        a = A()
        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            await a.some_func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertTrue(tests.error.wo_mandatory_location(str(violation_error)).startswith('self.x > 0'))



# pylint: disable=missing-docstring
# pylint: disable=no-self-use
# pylint: disable=invalid-name
# pylint: disable=unused-argument
# pylint: disable=no-member
# pylint: disable=unnecessary-lambda
import copy
import textwrap
import unittest
from typing import Optional, Any, Tuple, Dict

import icontract
import tests.error


class TestArgs(unittest.IsolatedAsyncioTestCase):
    async def test_args_without_variable_positional_arguments(self) -> None:
        recorded_args = None  # type: Optional[Tuple[Any, ...]]

        def set_args(args: Tuple[Any, ...]) -> bool:
            nonlocal recorded_args
            recorded_args = copy.copy(args)
            return True

        @icontract.require(lambda _ARGS: set_args(_ARGS))
        async def some_func(x: int) -> None:
            pass

        await some_func(3)

        assert recorded_args is not None
        self.assertTupleEqual((3, ), recorded_args)

    async def test_args_with_named_and_variable_positional_arguments(self) -> None:
        recorded_args = None  # type: Optional[Tuple[Any, ...]]

        def set_args(args: Tuple[Any, ...]) -> bool:
            nonlocal recorded_args
            recorded_args = copy.copy(args)
            return True

        @icontract.require(lambda _ARGS: set_args(_ARGS))
        async def some_func(x: int, *args: Any) -> None:
            pass

        await some_func(3, 2)

        assert recorded_args is not None
        self.assertTupleEqual((3, 2), recorded_args)

    async def test_args_with_only_variable_positional_arguments(self) -> None:
        recorded_args = None  # type: Optional[Tuple[Any, ...]]

        def set_args(args: Tuple[Any, ...]) -> bool:
            nonlocal recorded_args
            recorded_args = copy.copy(args)
            return True

        @icontract.require(lambda _ARGS: set_args(_ARGS))
        async def some_func(*args: Any) -> None:
            pass

        await some_func(3, 2, 1)

        assert recorded_args is not None
        self.assertTupleEqual((3, 2, 1), recorded_args)

    async def test_args_with_uncommon_variable_positional_arguments(self) -> None:
        recorded_args = None  # type: Optional[Tuple[Any, ...]]

        def set_args(args: Tuple[Any, ...]) -> bool:
            nonlocal recorded_args
            recorded_args = copy.copy(args)
            return True

        @icontract.require(lambda _ARGS: set_args(_ARGS))
        async def some_func(*arguments: Any) -> None:
            pass

        await some_func(3, 2, 1, 0)

        assert recorded_args is not None
        self.assertTupleEqual((3, 2, 1, 0), recorded_args)

    async def test_fail(self) -> None:
        @icontract.require(lambda _ARGS: len(_ARGS) > 2)
        async def some_func(*args: Any) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            await some_func(3)
        except icontract.ViolationError as err:
            violation_error = err

        assert violation_error is not None
        self.assertEqual(
            textwrap.dedent('''\
                len(_ARGS) > 2:
                _ARGS was (3,)
                len(_ARGS) was 1'''), tests.error.wo_mandatory_location(str(violation_error)))


class TestKwargs(unittest.IsolatedAsyncioTestCase):
    async def test_kwargs_without_variable_keyword_arguments(self) -> None:
        recorded_kwargs = None  # type: Optional[Dict[str, Any]]

        def set_kwargs(kwargs: Dict[str, Any]) -> bool:
            nonlocal recorded_kwargs
            recorded_kwargs = copy.copy(kwargs)
            return True

        @icontract.require(lambda _KWARGS: set_kwargs(_KWARGS))
        async def some_func(x: int) -> None:
            pass

        await some_func(x=3)

        assert recorded_kwargs is not None
        self.assertDictEqual({"x": 3}, recorded_kwargs)

    async def test_kwargs_with_named_and_variable_keyword_arguments(self) -> None:
        recorded_kwargs = None  # type: Optional[Dict[str, Any]]

        def set_kwargs(kwargs: Dict[str, Any]) -> bool:
            nonlocal recorded_kwargs
            recorded_kwargs = copy.copy(kwargs)
            return True

        @icontract.require(lambda _KWARGS: set_kwargs(_KWARGS))
        async def some_func(x: int, **kwargs: Any) -> None:
            pass

        await some_func(x=3, y=2)

        assert recorded_kwargs is not None
        self.assertDictEqual({"x": 3, "y": 2}, recorded_kwargs)

    async def test_kwargs_with_only_variable_keyword_arguments(self) -> None:
        recorded_kwargs = None  # type: Optional[Dict[str, Any]]

        def set_kwargs(kwargs: Dict[str, Any]) -> bool:
            nonlocal recorded_kwargs
            recorded_kwargs = copy.copy(kwargs)
            return True

        @icontract.require(lambda _KWARGS: set_kwargs(_KWARGS))
        async def some_func(**kwargs: Any) -> None:
            pass

        await some_func(x=3, y=2, z=1)

        assert recorded_kwargs is not None
        self.assertDictEqual({"x": 3, "y": 2, "z": 1}, recorded_kwargs)

    async def test_kwargs_with_uncommon_argument_name_for_variable_keyword_arguments(self) -> None:
        recorded_kwargs = None  # type: Optional[Dict[str, Any]]

        def set_kwargs(kwargs: Dict[str, Any]) -> bool:
            nonlocal recorded_kwargs
            recorded_kwargs = copy.copy(kwargs)
            return True

        @icontract.require(lambda _KWARGS: set_kwargs(_KWARGS))
        async def some_func(**parameters: Any) -> None:
            pass

        await some_func(x=3, y=2, z=1, a=0)

        assert recorded_kwargs is not None
        self.assertDictEqual({"x": 3, "y": 2, "z": 1, "a": 0}, recorded_kwargs)

    async def test_fail(self) -> None:
        @icontract.require(lambda _KWARGS: 'x' in _KWARGS)
        async def some_func(**kwargs: Any) -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            await some_func(y=3)
        except icontract.ViolationError as err:
            violation_error = err

        assert violation_error is not None
        self.assertEqual(
            textwrap.dedent("'x' in _KWARGS: _KWARGS was {'y': 3}"),
            tests.error.wo_mandatory_location(str(violation_error)))


class TestArgsAndKwargs(unittest.IsolatedAsyncioTestCase):
    async def test_that_args_and_kwargs_are_both_passed_as_placeholders(self) -> None:
        recorded_args = None  # type: Optional[Tuple[Any, ...]]
        recorded_kwargs = None  # type: Optional[Dict[str, Any]]

        def set_args_and_kwargs(args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> bool:
            nonlocal recorded_args
            nonlocal recorded_kwargs
            recorded_args = copy.copy(args)
            recorded_kwargs = copy.copy(kwargs)
            return True

        @icontract.require(lambda _ARGS, _KWARGS: set_args_and_kwargs(_ARGS, _KWARGS))
        async def some_func(*args: Any, **kwargs: Any) -> None:
            pass

        await some_func(5, x=10, y=20, z=30)

        assert recorded_args is not None
        self.assertTupleEqual((5, ), recorded_args)

        assert recorded_kwargs is not None
        self.assertDictEqual({"x": 10, "y": 20, "z": 30}, recorded_kwargs)

    async def test_a_very_mixed_case(self) -> None:
        recorded_args = None  # type: Optional[Tuple[Any, ...]]
        recorded_kwargs = None  # type: Optional[Dict[str, Any]]

        def set_args_and_kwargs(args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> bool:
            nonlocal recorded_args
            nonlocal recorded_kwargs
            recorded_args = copy.copy(args)
            recorded_kwargs = copy.copy(kwargs)
            return True

        @icontract.require(lambda _ARGS, _KWARGS: set_args_and_kwargs(_ARGS, _KWARGS))
        async def some_func(x: int, y: int, *args: Any, **kwargs: Any) -> None:
            pass

        await some_func(5, 10, 20, z=30)

        assert recorded_args is not None
        self.assertTupleEqual((5, 10, 20), recorded_args)

        assert recorded_kwargs is not None
        self.assertDictEqual({"z": 30}, recorded_kwargs)


class TestConflictOnARGSReported(unittest.IsolatedAsyncioTestCase):
    async def test_in_kwargs_in_call(self) -> None:
        @icontract.require(lambda _ARGS: True)
        async def some_func(*args, **kwargs) -> None:  # type: ignore
            pass

        type_error = None  # type: Optional[TypeError]
        try:
            await some_func(_ARGS="something")
        except TypeError as error:
            type_error = error

        assert type_error is not None
        self.assertEqual('The arguments of the function call include "_ARGS" '
                         'which is a placeholder for positional arguments in a condition.', str(type_error))


class TestConflictOnKWARGSReported(unittest.IsolatedAsyncioTestCase):
    async def test_in_kwargs_in_call(self) -> None:
        @icontract.require(lambda _ARGS: True)
        async def some_func(*args, **kwargs) -> None:  # type: ignore
            pass

        type_error = None  # type: Optional[TypeError]
        try:
            await some_func(_KWARGS="something")
        except TypeError as error:
            type_error = error

        assert type_error is not None
        self.assertEqual('The arguments of the function call include "_KWARGS" '
                         'which is a placeholder for keyword arguments in a condition.', str(type_error))


if __name__ == '__main__':
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=no-member
# pylint: disable=unused-argument
# pylint: disable=unnecessary-lambda

import unittest
from typing import Optional, List

import icontract

import tests.error
import tests.mock


class TestAsyncFunctionSyncCondition(unittest.IsolatedAsyncioTestCase):
    async def test_ok(self) -> None:
        @icontract.ensure(lambda result: result > 0)
        async def some_func() -> int:
            return 100

        result = await some_func()
        self.assertEqual(100, result)

    async def test_ok_with_snapshot(self) -> None:
        @icontract.snapshot(lambda lst: len(lst), name="len_lst")
        @icontract.ensure(lambda OLD, result: result > OLD.len_lst)
        async def some_func(lst: List[int]) -> int:
            lst.append(1984)
            return len(lst)

        result = await some_func(lst=[1, 2, 3])
        self.assertEqual(4, result)

    async def test_fail(self) -> None:
        @icontract.ensure(lambda result: result > 0)
        async def some_func() -> int:
            return -100

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = await some_func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual("result > 0: result was -100", tests.error.wo_mandatory_location(str(violation_error)))


class TestAsyncFunctionAsyncCondition(unittest.IsolatedAsyncioTestCase):
    async def test_ok(self) -> None:
        async def result_greater_zero(result: int) -> bool:
            return result > 0

        @icontract.ensure(result_greater_zero)
        async def some_func() -> int:
            return 100

        result = await some_func()
        self.assertEqual(100, result)

    async def test_snapshot(self) -> None:
        async def capture_len_lst(lst: List[int]) -> int:
            return len(lst)

        @icontract.snapshot(capture_len_lst, name="len_lst")
        @icontract.ensure(lambda OLD, lst: OLD.len_lst + 1 == len(lst))
        async def some_func(lst: List[int]) -> None:
            lst.append(1984)

        lst = []  # type: List[int]
        await some_func(lst=lst)
        self.assertListEqual([1984], lst)

    async def test_fail(self) -> None:
        async def result_greater_zero(result: int) -> bool:
            return result > 0

        @icontract.ensure(result_greater_zero)
        async def some_func() -> int:
            return -100

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = await some_func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual("result_greater_zero: result was -100", tests.error.wo_mandatory_location(
            str(violation_error)))


class TestCoroutine(unittest.IsolatedAsyncioTestCase):
    async def test_ok(self) -> None:
        async def some_condition() -> bool:
            return True

        @icontract.ensure(lambda: some_condition())
        async def some_func() -> None:
            pass

        await some_func()

    async def test_fail(self) -> None:
        async def some_condition() -> bool:
            return False

        @icontract.ensure(lambda: some_condition(), error=lambda: icontract.ViolationError("hihi"))
        async def some_func() -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            await some_func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual("hihi", str(violation_error))

    async def test_reported_if_no_error_is_specified_as_we_can_not_recompute_coroutine_functions(self) -> None:
        async def some_condition() -> bool:
            return False

        @icontract.ensure(lambda: some_condition())
        async def some_func() -> None:
            pass

        runtime_error = None  # type: Optional[RuntimeError]
        try:
            await some_func()
        except RuntimeError as err:
            runtime_error = err

        assert runtime_error is not None
        assert runtime_error.__cause__ is not None
        assert isinstance(runtime_error.__cause__, ValueError)

        value_error = runtime_error.__cause__

        self.assertRegex(
            str(value_error), r"^Unexpected coroutine function <function .*> as a condition of a contract\. "
            r"You must specify your own error if the condition of your contract is a coroutine function\.")

    async def test_snapshot(self) -> None:
        async def some_capture() -> int:
            return 1984

        @icontract.snapshot(lambda: some_capture(), name="hoho")
        @icontract.ensure(lambda OLD: OLD.hoho == 1984)
        async def some_func() -> None:
            pass

        await some_func()


class TestInvalid(unittest.IsolatedAsyncioTestCase):
    async def test_invalid_postcondition_arguments(self) -> None:
        @icontract.ensure(lambda b, result: b > result)
        async def some_function(a: int) -> None:  # pylint: disable=unused-variable
            pass

        type_err = None  # type: Optional[TypeError]
        try:
            await some_function(a=13)
        except TypeError as err:
            type_err = err

        self.assertIsNotNone(type_err)
        self.assertEqual("The argument(s) of the contract condition have not been set: ['b']. "
                         "Does the original function define them? Did you supply them in the call?",
                         tests.error.wo_mandatory_location(str(type_err)))

    async def test_conflicting_result_argument(self) -> None:
        @icontract.ensure(lambda a, result: a > result)
        async def some_function(a: int, result: int) -> None:  # pylint: disable=unused-variable
            pass

        type_err = None  # type: Optional[TypeError]
        try:
            await some_function(a=13, result=2)
        except TypeError as err:
            type_err = err

        self.assertIsNotNone(type_err)
        self.assertEqual("Unexpected argument 'result' in a function decorated with postconditions.", str(type_err))

    async def test_conflicting_OLD_argument(self) -> None:
        @icontract.snapshot(lambda a: a[:])
        @icontract.ensure(lambda OLD, a: a == OLD.a)
        async def some_function(a: List[int], OLD: int) -> None:  # pylint: disable=unused-variable
            pass

        type_err = None  # type: Optional[TypeError]
        try:
            await some_function(a=[13], OLD=2)
        except TypeError as err:
            type_err = err

        self.assertIsNotNone(type_err)
        self.assertEqual("Unexpected argument 'OLD' in a function decorated with postconditions.", str(type_err))

    async def test_error_with_invalid_arguments(self) -> None:
        @icontract.ensure(
            lambda result: result > 0, error=lambda z, result: ValueError("x is {}, result is {}".format(z, result)))
        async def some_func(x: int) -> int:
            return x

        type_error = None  # type: Optional[TypeError]
        try:
            await some_func(x=0)
        except TypeError as err:
            type_error = err

        self.assertIsNotNone(type_error)
        self.assertEqual("The argument(s) of the contract error have not been set: ['z']. "
                         "Does the original function define them? Did you supply them in the call?",
                         tests.error.wo_mandatory_location(str(type_error)))

    async def test_no_boolyness(self) -> None:
        @icontract.ensure(lambda: tests.mock.NumpyArray([True, False]))
        async def some_func() -> None:
            pass

        value_error = None  # type: Optional[ValueError]
        try:
            await some_func()
        except ValueError as err:
            value_error = err

        self.assertIsNotNone(value_error)
        self.assertEqual('Failed to negate the evaluation of the condition.',
                         tests.error.wo_mandatory_location(str(value_error)))


if __name__ == '__main__':
    unittest.main()



# pylint: disable=missing-docstring
# pylint: disable=invalid-name
# pylint: disable=no-member
# pylint: disable=unused-argument
# pylint: disable=unnecessary-lambda

import unittest
from typing import Optional

import icontract
import tests.error


class TestAsyncFunctionSyncCondition(unittest.IsolatedAsyncioTestCase):
    async def test_ok(self) -> None:
        @icontract.require(lambda x: x > 0)
        async def some_func(x: int) -> int:
            return x * 10

        result = await some_func(1)
        self.assertEqual(10, result)

    async def test_fail(self) -> None:
        @icontract.require(lambda x: x > 0)
        async def some_func(x: int) -> int:
            return x * 10

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = await some_func(-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual("x > 0: x was -1", tests.error.wo_mandatory_location(str(violation_error)))


class TestAsyncFunctionAsyncCondition(unittest.IsolatedAsyncioTestCase):
    async def test_ok(self) -> None:
        async def x_greater_zero(x: int) -> bool:
            return x > 0

        @icontract.require(x_greater_zero)
        async def some_func(x: int) -> int:
            return x * 10

        result = await some_func(1)
        self.assertEqual(10, result)

    async def test_fail(self) -> None:
        async def x_greater_zero(x: int) -> bool:
            return x > 0

        @icontract.require(x_greater_zero)
        async def some_func(x: int) -> int:
            return x * 10

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            _ = await some_func(-1)
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual("x_greater_zero: x was -1", tests.error.wo_mandatory_location(str(violation_error)))


class TestCoroutine(unittest.IsolatedAsyncioTestCase):
    async def test_ok(self) -> None:
        async def some_condition() -> bool:
            return True

        @icontract.require(lambda: some_condition())
        async def some_func() -> None:
            pass

        await some_func()

    async def test_fail(self) -> None:
        async def some_condition() -> bool:
            return False

        @icontract.require(lambda: some_condition(), error=lambda: icontract.ViolationError("hihi"))
        async def some_func() -> None:
            pass

        violation_error = None  # type: Optional[icontract.ViolationError]
        try:
            await some_func()
        except icontract.ViolationError as err:
            violation_error = err

        self.assertIsNotNone(violation_error)
        self.assertEqual("hihi", str(violation_error))

    async def test_reported_if_no_error_is_specified_as_we_can_not_recompute_coroutine_functions(self) -> None:
        async def some_condition() -> bool:
            return False

        @icontract.require(lambda: some_condition())
        async def some_func() -> None:
            pass

        runtime_error = None  # type: Optional[RuntimeError]
        try:
            await some_func()
        except RuntimeError as err:
            runtime_error = err

        assert runtime_error is not None
        assert runtime_error.__cause__ is not None
        assert isinstance(runtime_error.__cause__, ValueError)

        value_error = runtime_error.__cause__

        self.assertRegex(
            str(value_error), r"^Unexpected coroutine function <function .*> as a condition of a contract\. "
            r"You must specify your own error if the condition of your contract is a coroutine function\.")


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring

import dataclasses
import unittest
from typing import List

import asyncstdlib as a

import icontract


class TestCoroutines(unittest.IsolatedAsyncioTestCase):
    async def test_mock_backend_example(self) -> None:
        # This is an example of a backend system.
        # This test demonstrates how contracts can be used with coroutines.

        async def has_author(identifier: str) -> bool:
            return identifier in ["Margaret Cavendish", "Jane Austen"]

        async def has_category(category: str) -> bool:
            return category in await get_categories()

        async def get_categories() -> List[str]:
            return ["sci-fi", "romance"]

        @dataclasses.dataclass
        class Book:
            identifier: str
            author: str

        @icontract.require(lambda categories: a.map(has_category, categories))
        @icontract.ensure(lambda result: a.all(a.await_each(has_author(book.author) for book in result)))
        async def list_books(categories: List[str]) -> List[Book]:
            result = []  # type: List[Book]
            for category in categories:
                if category == "sci-fi":
                    result.extend([Book(identifier="The Blazing World", author="Margaret Cavendish")])
                elif category == "romance":
                    result.extend([Book(identifier="Pride and Prejudice", author="Jane Austen")])
                else:
                    raise AssertionError(category)

            return result

        sci_fi_books = await list_books(categories=['sci-fi'])
        self.assertListEqual(['The Blazing World'], [book.identifier for book in sci_fi_books])



"""Handle recomputation of values of a function given its abstract syntax tree and function frame."""

import ast
import builtins
import functools
import inspect
import platform
import sys
from typing import Any, Mapping, Dict, List, Optional, Union, Tuple, Set, Callable, \
    cast  # pylint: disable=unused-import


class Placeholder:
    """Represent a placeholder for variables local to the lambda such as targets in generator expressions."""

    def __repr__(self) -> str:
        """Represent the placeholder as <Placeholder>."""
        return "<Placeholder>"


PLACEHOLDER = Placeholder()


class Visitor(ast.NodeVisitor):
    """
    Traverse the abstract syntax tree and recompute the values of each node defined by the function frame.

    :ivar recomputed_values: mapping node -> value assigned to each visited node
    :type recomputed_values: Mapping[ast.AST, Any]
    """

    # pylint: disable=invalid-name
    # pylint: disable=missing-docstring
    # pylint: disable=too-many-public-methods

    def __init__(self, variable_lookup: List[Mapping[str, Any]]) -> None:
        """
        Initialize.

        :param variable_lookup: list of lookup tables to look-up the values of the variables, sorted by precedence
        """
        # _name_to_value maps the variable names to variable values.
        # This is important for Load contexts as well as Store contexts in, e.g., named expressions.
        self._name_to_value = dict()  # type: Dict[str, Any]

        # Resolve precedence of variable lookups
        for lookup in variable_lookup:
            for name, value in lookup.items():
                if name not in self._name_to_value:
                    self._name_to_value[name] = value

        # value assigned to each visited node
        self.recomputed_values = dict()  # type: Dict[ast.AST, Any]

    # pylint: disable=no-member
    if sys.version_info < (3, 8):

        def visit_Num(self, node: ast.Num) -> Union[int, float]:
            """Recompute the value as the number at the node."""
            result = node.n

            self.recomputed_values[node] = result

            assert isinstance(result, (int, float))
            return result

        def visit_Str(self, node: ast.Str) -> str:
            """Recompute the value as the string at the node."""
            result = node.s

            self.recomputed_values[node] = result
            return result

        def visit_Bytes(self, node: ast.Bytes) -> bytes:
            """Recompute the value as the bytes at the node."""
            result = node.s

            self.recomputed_values[node] = result
            return node.s

        def visit_NameConstant(self, node: ast.NameConstant) -> Any:
            """Forward the node value as a result."""
            self.recomputed_values[node] = node.value
            return node.value
    else:

        def visit_Constant(self, node: ast.Constant) -> Any:
            """Forward the node value as a result."""
            self.recomputed_values[node] = node.value
            return node.value

    if sys.version_info >= (3, 6):

        def visit_FormattedValue(self, node: ast.FormattedValue) -> Any:
            """Format the node value."""
            fmt = ['{']
            # See https://docs.python.org/3/library/ast.html#ast.FormattedValue for these
            # constants
            if node.conversion == -1:
                pass
            elif node.conversion == 115:
                fmt.append('!s')
            elif node.conversion == 114:
                fmt.append('!r')
            elif node.conversion == 97:
                fmt.append('!a')
            else:
                raise NotImplementedError("Unhandled conversion of a formatted value node {!r}: {}".format(
                    node, node.conversion))

            if node.format_spec is not None:
                fmt.append(":")

                # The following assert serves only documentation purposes so that the code is easier to follow.
                assert isinstance(node.format_spec, ast.JoinedStr)
                fmt.append(self.visit(node.format_spec))

            fmt.append('}')

            recomputed_value = self.visit(node.value)
            return ''.join(fmt).format(recomputed_value)

        def visit_JoinedStr(self, node: ast.JoinedStr) -> Any:
            """Visit the values and concatenate them."""
            joined_str = ''.join(self.visit(value_node) for value_node in node.values)

            self.recomputed_values[node] = joined_str
            return joined_str

    # pylint: enable=no-member

    def visit_List(self, node: ast.List) -> List[Any]:
        """Visit the elements and assemble the results into a list."""
        if isinstance(node.ctx, ast.Store):
            raise NotImplementedError("Can not compute the value of a Store on a list")

        result = [self.visit(node=elt) for elt in node.elts]

        self.recomputed_values[node] = result
        return result

    def visit_Tuple(self, node: ast.Tuple) -> Tuple[Any, ...]:
        """Visit the elements and assemble the results into a tuple."""
        if isinstance(node.ctx, ast.Store):
            raise NotImplementedError("Can not compute the value of a Store on a tuple")

        result = tuple(self.visit(node=elt) for elt in node.elts)

        self.recomputed_values[node] = result
        return result

    def visit_Set(self, node: ast.Set) -> Set[Any]:
        """Visit the elements and assemble the results into a set."""
        result = set(self.visit(node=elt) for elt in node.elts)

        self.recomputed_values[node] = result
        return result

    def visit_Dict(self, node: ast.Dict) -> Dict[Any, Any]:
        """Visit keys and values and assemble a dictionary with the results."""
        recomputed_dict = dict()  # type: Dict[Any, Any]
        for key, val in zip(node.keys, node.values):
            assert isinstance(key, ast.AST)
            assert isinstance(val, ast.AST)

            recomputed_dict[self.visit(node=key)] = self.visit(node=val)

        self.recomputed_values[node] = recomputed_dict
        return recomputed_dict

    def visit_Name(self, node: ast.Name) -> Any:
        """Load the variable by looking it up in the variable look-up and in the built-ins."""
        if not isinstance(node.ctx, ast.Load):
            raise NotImplementedError("Can only compute a value of Load on a name {}, but got context: {}".format(
                node.id, node.ctx))

        result = None  # type: Optional[Any]

        if node.id in self._name_to_value:
            result = self._name_to_value[node.id]

        if result is None and hasattr(builtins, node.id):
            result = getattr(builtins, node.id)

        if result is None and node.id != "None":
            # The variable refers to a name local of the lambda (e.g., a target in the generator expression).
            # Since we evaluate generator expressions with runtime compilation, None is returned here as a placeholder.
            return PLACEHOLDER

        self.recomputed_values[node] = result
        return result

    def visit_Expr(self, node: ast.Expr) -> Any:
        """Visit the node's ``value``."""
        result = self.visit(node=node.value)

        self.recomputed_values[node] = result
        return result

    def visit_UnaryOp(self, node: ast.UnaryOp) -> Any:
        """Visit the node operand and apply the operation on the result."""
        if isinstance(node.op, ast.UAdd):
            result = +self.visit(node=node.operand)
        elif isinstance(node.op, ast.USub):
            result = -self.visit(node=node.operand)
        elif isinstance(node.op, ast.Not):
            result = not self.visit(node=node.operand)
        elif isinstance(node.op, ast.Invert):
            result = ~self.visit(node=node.operand)
        else:
            raise NotImplementedError("Unhandled op of {}: {}".format(node, node.op))

        self.recomputed_values[node] = result
        return result

    def visit_BinOp(self, node: ast.BinOp) -> Any:
        """Recursively visit the left and right operand, respectively, and apply the operation on the results."""
        # pylint: disable=too-many-branches
        left = self.visit(node=node.left)
        right = self.visit(node=node.right)

        if isinstance(node.op, ast.Add):
            result = left + right
        elif isinstance(node.op, ast.Sub):
            result = left - right
        elif isinstance(node.op, ast.Mult):
            result = left * right
        elif isinstance(node.op, ast.Div):
            result = left / right
        elif isinstance(node.op, ast.FloorDiv):
            result = left // right
        elif isinstance(node.op, ast.Mod):
            result = left % right
        elif isinstance(node.op, ast.Pow):
            result = left**right
        elif isinstance(node.op, ast.LShift):
            result = left << right
        elif isinstance(node.op, ast.RShift):
            result = left >> right
        elif isinstance(node.op, ast.BitOr):
            result = left | right
        elif isinstance(node.op, ast.BitXor):
            result = left ^ right
        elif isinstance(node.op, ast.BitAnd):
            result = left & right
        elif isinstance(node.op, ast.MatMult):
            result = left @ right
        else:
            raise NotImplementedError("Unhandled op of {}: {}".format(node, node.op))

        self.recomputed_values[node] = result
        return result

    def visit_BoolOp(self, node: ast.BoolOp) -> Any:
        """Recursively visit the operands and apply the operation on them."""
        values = [self.visit(value_node) for value_node in node.values]

        if isinstance(node.op, ast.And):
            result = functools.reduce(lambda left, right: left and right, values, True)
        elif isinstance(node.op, ast.Or):
            result = functools.reduce(lambda left, right: left or right, values, True)
        else:
            raise NotImplementedError("Unhandled op of {}: {}".format(node, node.op))

        self.recomputed_values[node] = result
        return result

    def visit_Compare(self, node: ast.Compare) -> Any:
        """Recursively visit the comparators and apply the operations on them."""
        # pylint: disable=too-many-branches
        left = self.visit(node=node.left)

        comparators = [self.visit(node=comparator) for comparator in node.comparators]

        result = None  # type: Optional[Any]
        for comparator, op in zip(comparators, node.ops):
            if isinstance(op, ast.Eq):
                comparison = left == comparator
            elif isinstance(op, ast.NotEq):
                comparison = left != comparator
            elif isinstance(op, ast.Lt):
                comparison = left < comparator
            elif isinstance(op, ast.LtE):
                comparison = left <= comparator
            elif isinstance(op, ast.Gt):
                comparison = left > comparator
            elif isinstance(op, ast.GtE):
                comparison = left >= comparator
            elif isinstance(op, ast.Is):
                comparison = left is comparator
            elif isinstance(op, ast.IsNot):
                comparison = left is not comparator
            elif isinstance(op, ast.In):
                comparison = left in comparator
            elif isinstance(op, ast.NotIn):
                comparison = left not in comparator
            else:
                raise NotImplementedError("Unhandled op of {}: {}".format(node, op))

            if result is None:
                result = comparison
            else:
                result = result and comparison

            left = comparator

        self.recomputed_values[node] = result
        return result

    def visit_Call(self, node: ast.Call) -> Any:
        """Visit the function and the arguments and finally make the function call with them."""
        func = self.visit(node=node.func)

        if not callable(func):
            raise ValueError(("Unexpected call to a non-callle during the re-computation: {}").format(func))

        if inspect.iscoroutinefunction(func):
            raise ValueError(
                ("Unexpected coroutine function {} as a condition of a contract. "
                 "You must specify your own error if the condition of your contract is a coroutine function."
                 ).format(func))

        args = []  # type: List[Any]
        for arg_node in node.args:
            if isinstance(arg_node, ast.Starred):
                args.extend(self.visit(node=arg_node))
            else:
                args.append(self.visit(node=arg_node))

        kwargs = dict()  # type: Dict[str, Any]
        for keyword in node.keywords:
            if keyword.arg is None:
                kw = self.visit(node=keyword.value)
                for key, val in kw.items():
                    kwargs[key] = val

            else:
                kwargs[keyword.arg] = self.visit(node=keyword.value)

        # If any of the positional or keyword arguments are placeholders, that means that we are re-computing a
        # generator expression.
        # As we re-compute them by re-compilation, we do not re-compute the individual calls here.
        if PLACEHOLDER in args or PLACEHOLDER in kwargs.values():
            return PLACEHOLDER

        result = func(*args, **kwargs)

        if inspect.iscoroutine(result):
            raise ValueError(
                ("Unexpected coroutine {} as a result from a call. "
                 "You must specify your own error if the condition of your contract gives a coroutine.").format(result))

        self.recomputed_values[node] = result
        return result

    def visit_IfExp(self, node: ast.IfExp) -> Any:
        """Visit the ``test``, and depending on its outcome, the ``body`` or ``orelse``."""
        test = self.visit(node=node.test)

        if test:
            result = self.visit(node=node.body)
        else:
            result = self.visit(node=node.orelse)

        self.recomputed_values[node] = result
        return result

    def visit_Attribute(self, node: ast.Attribute) -> Any:
        """Visit the node's ``value`` and get the attribute from the result."""
        value = self.visit(node=node.value)
        if not isinstance(node.ctx, ast.Load):
            raise NotImplementedError(
                "Can only compute a value of Load on the attribute {}, but got context: {}".format(node.attr, node.ctx))

        result = getattr(value, node.attr)

        self.recomputed_values[node] = result
        return result

    if sys.version_info >= (3, 8):
        # pylint: disable=no-member
        def visit_NamedExpr(self, node: ast.NamedExpr) -> Any:
            """Visit the node's ``value`` and assign it to both this node and the target."""
            value = self.visit(node=node.value)
            self.recomputed_values[node] = value

            # This assignment is needed to make mypy happy.
            target = cast(ast.Name, node.target)

            if not isinstance(target.ctx, ast.Store):
                raise NotImplementedError(
                    "Expected Store context in the target of a named expression, but got: {}".format(target.ctx))

            self._name_to_value[target.id] = value

            return value

    def visit_Index(self, node: ast.Index) -> Any:
        """Visit the node's ``value``."""
        result = self.visit(node=node.value)

        self.recomputed_values[node] = result
        return result

    def visit_Slice(self, node: ast.Slice) -> slice:
        """Visit ``lower``, ``upper`` and ``step`` and recompute the node as a ``slice``."""
        lower = None  # type: Optional[int]
        if node.lower is not None:
            lower = self.visit(node=node.lower)

        upper = None  # type: Optional[int]
        if node.upper is not None:
            upper = self.visit(node=node.upper)

        step = None  # type: Optional[int]
        if node.step is not None:
            step = self.visit(node=node.step)

        result = slice(lower, upper, step)

        self.recomputed_values[node] = result
        return result

    def visit_ExtSlice(self, node: ast.ExtSlice) -> Tuple[Any, ...]:
        """Visit each dimension of the advanced slicing and assemble the dimensions in a tuple."""
        result = tuple(self.visit(node=dim) for dim in node.dims)

        self.recomputed_values[node] = result
        return result

    def visit_Subscript(self, node: ast.Subscript) -> Any:
        """Visit the ``slice`` and a ``value`` and get the element."""
        value = self.visit(node=node.value)
        a_slice = self.visit(node=node.slice)

        result = value[a_slice]

        self.recomputed_values[node] = result
        return result

    def _execute_comprehension(self, node: Union[ast.ListComp, ast.SetComp, ast.GeneratorExp, ast.DictComp]) -> Any:
        """Compile the generator or comprehension from the node and execute the compiled code."""
        args = [ast.arg(arg=name) for name in sorted(self._name_to_value.keys())]

        if platform.python_version_tuple() < ('3', ):
            raise NotImplementedError("Python versions below not supported, got: {}".format(platform.python_version()))

        if platform.python_version_tuple() < ('3', '8'):
            func_def_node = ast.FunctionDef(
                name="generator_expr",
                args=ast.arguments(args=args, kwonlyargs=[], kw_defaults=[], defaults=[]),
                decorator_list=[],
                body=[ast.Return(node)])

            module_node = ast.Module(body=[func_def_node])
        else:
            func_def_node = ast.FunctionDef(
                name="generator_expr",
                args=ast.arguments(args=args, posonlyargs=[], kwonlyargs=[], kw_defaults=[], defaults=[]),
                decorator_list=[],
                body=[ast.Return(node)])

            module_node = ast.Module(body=[func_def_node], type_ignores=[])

        ast.fix_missing_locations(module_node)

        code = compile(source=module_node, filename='<ast>', mode='exec')

        module_locals = {}  # type: Dict[str, Any]
        module_globals = {}  # type: Dict[str, Any]
        exec(code, module_globals, module_locals)  # pylint: disable=exec-used

        generator_expr_func = module_locals["generator_expr"]

        return generator_expr_func(**self._name_to_value)

    def visit_GeneratorExp(self, node: ast.GeneratorExp) -> Any:
        """Compile the generator expression as a function and call it."""
        result = self._execute_comprehension(node=node)

        for generator in node.generators:
            self.visit(generator.iter)

        # Do not set the computed value of the node since its representation would be non-informative.
        return result

    def visit_ListComp(self, node: ast.ListComp) -> Any:
        """Compile the list comprehension as a function and call it."""
        result = self._execute_comprehension(node=node)

        for generator in node.generators:
            self.visit(generator.iter)

        self.recomputed_values[node] = result
        return result

    def visit_SetComp(self, node: ast.SetComp) -> Any:
        """Compile the set comprehension as a function and call it."""
        result = self._execute_comprehension(node=node)

        for generator in node.generators:
            self.visit(generator.iter)

        self.recomputed_values[node] = result
        return result

    def visit_DictComp(self, node: ast.DictComp) -> Any:
        """Compile the dictionary comprehension as a function and call it."""
        result = self._execute_comprehension(node=node)

        for generator in node.generators:
            self.visit(generator.iter)

        self.recomputed_values[node] = result
        return result

    def visit_Lambda(self, node: ast.Lambda) -> Callable[..., Any]:
        """Do not support inline lambda until there is a feature request since this is quite tricky to implement."""
        raise NotImplementedError(
            "Recomputation of in-line lambda functions is not supported since it is quite tricky to implement and "
            "we decided to implement it only once there is a real need for it. "
            "Please make a feature request on https://github.com/Parquery/icontract")

    def visit_Return(self, node: ast.Return) -> Any:  # pylint: disable=no-self-use
        """Raise an exception that this node is unexpected."""
        raise AssertionError("Unexpected return node during the re-computation: {}".format(ast.dump(node)))

    def generic_visit(self, node: ast.AST) -> None:
        """Raise an exception that this node has not been handled."""
        raise NotImplementedError("Unhandled recomputation of the node: {} {}".format(type(node), node))
